#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "seg2_struct.h"
#include "externals.h"
#include "procs.h"
/* #include "qlib2.h" */

/************************************************************************
*  read_seg2_file_desc:							*
*	Read SEG2 file descriptor block,				*
*	determining whether the file's data is stored in my wordorder	*
*	or opposite wordorder.
*	Fill in file descriptor block structure stored in my wordorder.	*
*  return:								*
*	Number of traces in the SEG-Y file.				*
************************************************************************/
char read_seg2_file_desc
   (FILE	  *input,
    SEG2FILEDESC  *file_desc,
    int           *numTraces)
{
  size_t (*read_chars)(void *, size_t, size_t, FILE *); /* function pointer */
  int    numRead; /* Number of chars read from input, for error checking */
  int    i;     /* index to file_desc->trace_ptrs[] array */
  char   temp_chars_18[18];

  /***************************************************************
  * Read the SEG-2 File Descriptor's Block ID (First two bytes), *
  * determine if the format of the file is my wordorder or not , *
  * and set the file character reading function accordingly.     *
  ***************************************************************/
  numRead = fread(&file_desc->block_ID, sizeof(short int), 1, input);
  if (numRead != 1) {
    fprintf(stderr,
            "Error reading SEG-2 File Descriptor Block ID. # bytes read = %d, wanted 2; string read = %s.\n",
            numRead, file_desc->block_ID);
    return(1);
  }
  if      (file_desc->block_ID == NATIVE_WORDORDER_ID) { /* 0x3A55 */
    read_chars = fread; /* assign the function pointer */
  }
  else if (file_desc->block_ID == OPPOSITE_WORDORDER_ID) { /* 0x553A */
    read_chars = fread_swab; /* assign the function pointer */
  }
  else {
    fprintf(stderr,
            "Error: invalid SEG-2 File Descriptor Block ID: 0x%X. Skipping SEG-2 file.\n",
            file_desc->block_ID);
    return(1);
  }

  /************************************************
  * Read the next 12 bytes of the file descriptor *
  ************************************************/
  numRead = read_chars(&file_desc->revisionNum,          sizeof(short int), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: Revision Number",   "SEG-2 File" )) { return(1); }

  numRead = read_chars(&file_desc->tracePtrSubBlockSize, sizeof(short int), 1, input);
  if (readSEG2failed(numRead, 1,
                     "File Desc: Trace Ptr Sub Block Size",      "SEG-2 File" )) { return(1); }

  numRead = read_chars(&file_desc->numOfTraces,          sizeof(short int), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: Number of Traces",  "SEG-2 File" )) { return(1); }

  /* Single bytes can be read directly with fread (for speed): *
   * byte order (big/little endian) doesn't matter, obviously. */
  numRead = fread(&file_desc->strTermSize,  sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: String Term size",  "SEG-2 File" )) { return(1); }

  numRead = fread(&file_desc->strTerm1,     sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: String Term 1",     "SEG-2 File" )) { return(1); }

  numRead = fread(&file_desc->strTerm2,     sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: String Term 2",     "SEG-2 File" )) { return(1); }
    
  file_desc->strTerm[0] = file_desc->strTerm1;
  file_desc->strTerm[1] = (file_desc->strTermSize == 2) ? file_desc->strTerm2 : '\0';
  file_desc->strTerm[2] = '\0';

  numRead = fread(&file_desc->lineTermSize, sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: Line Term Size",    "SEG-2 File" )) { return(1); }

  numRead = fread(&file_desc->lineTerm1,    sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: Line Terminator 1", "SEG-2 File" )) { return(1); }

  numRead = fread(&file_desc->lineTerm2,    sizeof(char), 1, input);
  if (readSEG2failed(numRead, 1, "File Desc: Line Terminator 2", "SEG-2 File" )) { return(1); }

  /* Skip 18 reserved bytes (bytes 14 - 31) */
  numRead = fread(temp_chars_18, sizeof(char), 18, input);
  if (numRead != 18) {
    fprintf(stderr,
            "Error: while reading SEG2 File Descriptor, set of new file position failed on input %s.\n",
            *input);
    return(1);
  }

  /**************************************************************
  * Allocate space for an array of trace pointers, and fill     *
  * the array by reading each pointer from the file descriptor. *
  **************************************************************/
  *numTraces = file_desc->numOfTraces;
  file_desc->trace_ptrs = (int *)malloc(sizeof(int) * (*numTraces));
  for (i = 0; i < *numTraces; i++) {
    numRead = read_chars(&(file_desc->trace_ptrs)[i], sizeof(int), 1, input);
    if (numRead != 1) {
      fprintf(stderr,
              "Error: can't read SEG2's File Desc: Trace Pointer %d. Skipping SEG-2 File.\n",
              i);
      return(1);
    }
  }
 
  output_debug_fd_vals(file_desc);   /* output any debugging info requested */
  output_debug_fd_traces(file_desc); /* output any debugging info requested */

  /* Skip to the end of the trace pointer subblock. */
  if (fseeko (input, (off_t)32+file_desc->tracePtrSubBlockSize, SEEK_SET) != 0) {
    fprintf (stderr, "Error: seeking to file string descriptors\n");
   return(1);
  }

  /*****************************************************************************
  * Allocate space for an array of characters, and fill the array by reading   *
  * each character from the file descriptor.                                   *
  * The number of characters to read for the strings sub block is equal to     *
  *   the address of the first trace                                           *
  *   - (the size of the fixed part of the header [32 bytes]                   *
  *      + the size of the header's trace pointer sub block).                  *
  *****************************************************************************/
  file_desc->charsIn_strings = file_desc->trace_ptrs[0] - (32 + file_desc->tracePtrSubBlockSize);
  file_desc->strings = (char *)malloc(sizeof(char) * file_desc->charsIn_strings);
  numRead = fread(file_desc->strings, sizeof(char), file_desc->charsIn_strings, input);
  if (readSEG2failed(numRead, file_desc->charsIn_strings,
                     "File Desc: strings", "SEG-2 File" )) { return(1); }
  if (parse_strings_of_file_desc(file_desc) != 0) {
    fprintf(stderr, "Error parsing SEG-2 File Descriptor strings.  Skipping SEG-2 file.\n");
    return(1);
  }
  output_debug_fd_strs_parsed(file_desc); /* output any debugging info requested */

  return 0;
}


char parse_strings_of_file_desc(SEG2FILEDESC *file_desc)
{
  extern SEG2_KEYWORDS * file_desc_keywords; /* defined in seg2toms.c */
  int  max_possible_num_of_kv_pairs;
  KEY_VAL_PAIR **keys_and_vals;
  int  key_val_index = 0;
  int  index = 0;
  int  chars_to_next_pair;
  char return_val = 0;
  int  unknown_keys_index = 0; /* used when unknown key/value pairs are finally stored in file_desc. */
  int  unknown_keys_count = 0;
  int *key_categories;         /* number of elements will be key_val_index, once it is determined */

  max_possible_num_of_kv_pairs = file_desc->charsIn_strings / (2 + file_desc->strTermSize);
  keys_and_vals = (KEY_VAL_PAIR **)malloc (max_possible_num_of_kv_pairs * sizeof (KEY_VAL_PAIR *));
  if (keys_and_vals == NULL) {
    fprintf (stderr, "Error mallocing keys_and_vals\n");
    return(1);
  }

  /* Initiallize all the pointers of the known_keys to NULL */
  for ( key_val_index = 0; key_val_index < NUM_OF_FD_KEYWORDS; key_val_index++) {
    file_desc->known_keys[key_val_index] = NULL;
  }
  key_val_index = 0;

  keys_and_vals[key_val_index] = get_key_val_pair(file_desc, file_desc->strings, index, &chars_to_next_pair);
  while (keys_and_vals[key_val_index] != NULL) {
    index += chars_to_next_pair;
    key_val_index++;
    keys_and_vals[key_val_index] = get_key_val_pair(file_desc, file_desc->strings, index, &chars_to_next_pair);
  }

  key_categories = (int *)malloc(sizeof(int) * key_val_index);
  if (key_categories == NULL) {
    fprintf(stderr,
            "parse_strings_of_file_desc ERROR: Can't allocate memory block for int key_categories[%d].\n",
            key_val_index);
    return(1);
  }

  unknown_keys_count = categorize_keys(keys_and_vals, key_val_index, key_categories, file_desc_keywords);
  file_desc->unknown_keys_count = unknown_keys_count;

  if (unknown_keys_count > 0) {
    /* Allocate memory space for the unknown_keys array in file_desc */
    file_desc->unknown_keys = (KEY_VAL_PAIR **)malloc(sizeof(KEY_VAL_PAIR *) * unknown_keys_count);
    if (file_desc->unknown_keys == NULL) {
      fprintf (stderr,
               "parse_strings_of_file_desc ERROR: Can't allocate memory block for KEY_VAL_PAIR unknown_keys[%d].\n",
               unknown_keys_count);
      return(1);
    }
    /* Clear the unknown_keys array */
    memset ((void *)file_desc->unknown_keys, 0, sizeof(KEY_VAL_PAIR *) * unknown_keys_count);
  }

  /* Store the KEY_VAL_PAIRs found in the strings section of the            *
   * file descriptor block according to whether or not each key matched the *
   * known keywords (those listed in the SEG2 Specification, or those known *
   * at the time of programming: more can be added in seg2_struct.h).       */
  for (index = 0; index < key_val_index; index++) {
    if (key_categories[index] != -1) {
      file_desc->known_keys[key_categories[index]] = keys_and_vals[index];
    }
    else {
      file_desc->unknown_keys[unknown_keys_index++] = keys_and_vals[index];
    }
  }

  free(keys_and_vals);
  free(key_categories);

  return return_val;
}


void output_debug_fd_vals(SEG2FILEDESC *file_desc)
{
  extern int dbgflag; /* debug flag using DBG_FLAGS bit set, defined in externals.h */
  if (dbgflag & DBG_FLAG_FD_VALS) {
    fprintf(info,
            "File descriptor fields:\n"
            "  block_ID             = 0x%X\n"
            "    (0x3A55: my wordorder, 0x553A: opposite.\n"
            "     rest of data shown in my wordorder.)\n"
            "  revisionNum          = %d\n"
            "  tracePtrSubBlockSize = %d\n"
            "  numOfTraces          = %d\n"
            "  strTermSize          = %d\n"
            "  strTerm1             = %-2d (ASCII numerical representation)\n"
            "  strTerm2             = %-2d (ASCII numerical representation)\n"
            "  lineTermSize         = %d\n"
            "  lineTerm1            = %-2d (ASCII numerical representation)\n"
            "  lineTerm2            = %-2d (ASCII numerical representation)\n",
            file_desc->block_ID,
            file_desc->revisionNum,
            file_desc->tracePtrSubBlockSize,
            file_desc->numOfTraces,
            file_desc->strTermSize,
            file_desc->strTerm1,
            file_desc->strTerm2,
            file_desc->lineTermSize,
            file_desc->lineTerm1,
            file_desc->lineTerm2);
  }
}

void output_debug_fd_traces(SEG2FILEDESC *file_desc)
{
  extern int dbgflag; /* debug flag using DBG_FLAGS bit set, defined in externals.h */
  int i;

  if (dbgflag & DBG_FLAG_FD_TRACES) {
    fprintf(info,
            "File descriptor's %d trace pointers:\n",
            file_desc->numOfTraces);

    for (i = 0; i < file_desc->numOfTraces; i++) {
      fprintf(info,
              "  trace_ptrs[%03d] = %10d\n",
              i, (file_desc->trace_ptrs)[i]);
    }
  }
}

#if 0
void output_debug_fd_strs_whole(SEG2FILEDESC *file_desc)
{
  extern int dbgflag; /* debug flag using DBG_FLAGS bit set, defined in externals.h */
  if (dbgflag & DBG_FLAG_FD_STRS_WHOLE) {
    fprintf(info,
            "file descriptor strings are output here...  To be filled out\n");
  }
}
#endif

void output_debug_fd_strs_parsed(SEG2FILEDESC *file_desc)
{
  extern int dbgflag; /* debug flag using DBG_FLAGS bit set, defined in externals.h */
  int i;

  if (dbgflag & DBG_FLAG_FD_STRS_PARSED) {
    fprintf(info, "Expected File descriptor strings:\n");
    for (i = 0; i < NUM_OF_FD_KEYWORDS; i++) {
      if ((file_desc->known_keys)[i]) {
        fprintf(info,
                "  %30s = %s\n",
                file_desc->known_keys[i]->key,
                file_desc->known_keys[i]->val);
      }
    }
    if (file_desc->unknown_keys_count > 0) {
      fprintf(info, "Addition file descriptor strings:\n");
      for (i = 0; i < file_desc->unknown_keys_count; i++) {
        if ((file_desc->unknown_keys)[i]) {
          fprintf(info,
                  "  %30s = %s\n",
                  file_desc->unknown_keys[i]->key,
                  file_desc->unknown_keys[i]->val);
        }
      }
    }
  }
}


