#ifndef seg2_struct_h
#define seg2_struct_h

#include "qlib2.h"

/************************************************************************/
/* First 2 bytes of the file interpreted as a short int determines	*/
/* whether the file is in our native wordorder or not.			*/
/************************************************************************/
#define NATIVE_WORDORDER_ID	0x3A55
#define OPPOSITE_WORDORDER_ID	0x553A

typedef struct nscl_names {
  char station_id[DH_STATION_LEN+1];	/* station name		*/
  char location_id[DH_LOCATION_LEN+1];	/* location id		*/
  char channel_id[DH_CHANNEL_LEN+1];	/* channel name		*/
  char network_id[DH_NETWORK_LEN+1];	/* network id		*/
} NSCL;


/*******************************************************************************
*                    KEYWORD / VALUE PAIRS OF STRINGS SECTIONS:                *
* Definitions of keywords (and structures used to search for them in strings)  * 
* defined or known for SEG2 File and Trace Descriptor Blocks' strings section. *
*******************************************************************************/

/*
 * Each KEY_VAL_PAIR structure contains one keyword and its associated
 * value from the File and Trace Descriptor Blocks' string sections.
 */
typedef struct key_val_pair {
  int   key_length;
  int   val_length;
  char *key; /* array length to be determined when string is parsed, but will equal key_length. */
  char *val; /* array length to be determined when string is parsed, but will equal val_length. */
} KEY_VAL_PAIR;

/*
 * The SEG2_KEYWORDS structure is used to store the list of known keywords for
 * either the file descriptor block's string section, or
 * for  the  trace descriptor block's string section.
 * When it is being searched through to find a keyword, the keywords[] array is
 * treated like a circular buffer, the current_index indexes the array at
 * either the first keyword or at the first keyword after the last keyword
 * matched.  Since the keywords are stored in this sturcture in alphabetical
 * order, and they are supposed to be stored in alphabetical order in the SEG2
 * file and trace descriptors, this ensures the least number of searching will
 * be done to match the keywords, while still allowing for the keywords not to
 * be stored in the SEG2 file in alphabetical order.  The max index is used to
 * determine if the index has reached the end of the list and needs to be set
 * to zero in order to keep searching through all the words.  The current_index
 * is not reset unless the search for a keyword was successful.
 */
typedef struct seg2_keywords {
  int    current_index;
  int    max_index;
  char **keywords; /* gets assigned either the File or Trace Descriptor keywords */
} SEG2_KEYWORDS;

/*******  FILE DESCRIPTOR SPECIFIC KEYWORDS *******/
/*
 * The enumerations in fd_key_indices must be in the same order as
 * the strings list in the file_desc_keywords->keywords[] array.
 */
typedef enum fd_key_indices {
  ACQUISITION_DATE,
  ACQUISITION_SECOND_FRACTION,
  ACQUISITION_TIME,
  CLIENT,
  COMPANY,
  GENERAL_CONSTANT,
  INSTRUMENT,
  JOB_ID,
  OBSERVER,
  PROCESSING_DATE,
  PROCESSING_TIME,
  TRACE_SORT,
  UNITS,
  FD_NOTE,
  MAX_FD_KEY_INDEX = FD_NOTE,
  NUM_OF_FD_KEYWORDS,
} FD_KEY_INDICES;


/*******  TRACE DESCRIPTOR SPECIFIC KEYWORDS *******/
/*
 * The enumerations in td_key_indices must be in the same order as
 * the strings list in the trace_desc_keywords->keywords[] array.
 */
typedef enum td_key_indices {
  ALIAS_FILTER,
  AMPLITUDE_RECOVERY,
  BAND_REJECT_FILTER,
  CDP_NUMBER,
  CDP_TRACE,
  CHANNEL_NUMBER,
  DATUM,
  DELAY,
  DESCALING_FACTOR,
  DIGITAL_BAND_REJECT_FILTER,
  DIGITAL_LOW_CUT_FILTER,
  END_OF_GROUP,
  FIXED_GAIN,
  HIGH_CUT_FILTER,
  LINE_ID,
  LOW_CUT_FILTER,
  NOTCH_FREQUENCY,
  POLARITY,
  RAW_RECORD,
  RECEIVER,
  RECEIVER_GEOMETRY,
  RECEIVER_LOCATION,
  RECEIVER_SPECS,
  RECEIVER_STATION_NUMBER,
  SAMPLE_INTERVAL,
  SHOT_SEQUENCE_NUMBER,
  SKEW,
  SOURCE,
  SOURCE_LOCATION,
  SOURCE_STATION_NUMBER,
  STACK,
  STATIC_CORRECTIONS,
  TRACE_TYPE,
  TD_NOTE,
  MAX_TD_KEY_INDEX = TD_NOTE,
  NUM_OF_TD_KEYWORDS,
} TD_KEY_INDICES ;


/*********************************/
/*  SEG2 File Descriptor Block.  */
/*********************************/
typedef struct Seg2FileDescriptor {
  short int      block_ID;             /* 0x3A55: my endian, 0x553A: opposite endian */
  short int      revisionNum;          /*  */
  short int      tracePtrSubBlockSize; /*  */
  short int      numOfTraces;          /*  */
  char           strTermSize;          /*  */
  char           strTerm1;             /*  */
  char           strTerm2;             /*  */
  char           lineTermSize;         /*  */
  char           lineTerm1;            /*  */
  char           lineTerm2;            /*  */
  /* bytes 14 - 31 are reserved and unused (where the first byte is 0). */
  int           *trace_ptrs;           /*  */
  char          *strings;              /*  */
  /* The rest of this structure is not in the SEG2 specification */
  char           strTerm[3];           /* a string which combines strTerm1 and, if used, strTerm2. */
  unsigned int   charsIn_strings;      /* Size of the strings array. */
  int            unknown_keys_count;   /* count of key/value pairs whose keys didn't match known keys     */
  KEY_VAL_PAIR  *known_keys[NUM_OF_FD_KEYWORDS];
  KEY_VAL_PAIR **unknown_keys;         /* array size TBD and allocated when strings are parsed. */
  /* The strings[] array is parsed out using the SEG2 specification's
   * formatting rules, and each extracted keyword and value pair is stored
   * in a KEY_VAL_PAIR structure.  These pairs are then grouped into either
   * the known_keys[] or unknown_keys[], array based on whether each pair's
   * keyword was matched to one of the keywords in the array
   * file_desc_keywords->keywords[].
   */
} SEG2FILEDESC;


/**********************************/
/*  SEG2 Trace Descriptor Block.  */
/**********************************/
typedef struct Seg2TraceDescriptor {
  short int      block_ID;           /* 0x3A55: big endian, 0x553A: little endian */
  short int      traceDescBlockSize; /*  */
  int            traceDataBlockSize; /*  */
  int            numOfSamples;       /* NS: Number of Samples in Data Block */
  char           dataFormatCode;     /*  */
  /* bytes 13 - 31 are reserved and unused                      *
   * (where the first byte of the trace descriptor block is 0). */
  char          *strings;            /*  */
  /* The rest of this structure is not in the SEG2 specification */
  unsigned int   charsIn_strings;    /* Size of the strings array. */
  int            unknown_keys_count; /* count of key/value pairs whose keys didn't match known keys     */
  KEY_VAL_PAIR  *known_keys[NUM_OF_TD_KEYWORDS];
  KEY_VAL_PAIR **unknown_keys;       /* array size TBD and allocated when strings are parsed. */
  /* The strings[] array is parsed out using the SEG2 specification's
   * formatting rules, and each extracted keyword and value pair is stored
   * in a KEY_VAL_PAIR structure.  These pairs are then grouped into either
   * the known_keys[] or unknown_keys[], array based on whether each pair's
   * keyword was matched to one of the keywords in the array
   * trace_desc_keywords->keywords[].
   */
} SEG2TRACEDESC;

#endif

